'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var Joi = _interopDefault(require('@hapi/joi'));
var dynamodbAutoMarshaller = require('@aws/dynamodb-auto-marshaller');
var uuid = require('uuid');

class GenericDataError extends Error {
  constructor(message) {
    super(message);
  }

}
class GenericFriendlyError extends Error {
  constructor(message) {
    super(message);
  }

}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function ownKeys(object, enumerableOnly) {
  var keys = Object.keys(object);

  if (Object.getOwnPropertySymbols) {
    var symbols = Object.getOwnPropertySymbols(object);
    if (enumerableOnly) symbols = symbols.filter(function (sym) {
      return Object.getOwnPropertyDescriptor(object, sym).enumerable;
    });
    keys.push.apply(keys, symbols);
  }

  return keys;
}

function _objectSpread2(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};

    if (i % 2) {
      ownKeys(Object(source), true).forEach(function (key) {
        _defineProperty(target, key, source[key]);
      });
    } else if (Object.getOwnPropertyDescriptors) {
      Object.defineProperties(target, Object.getOwnPropertyDescriptors(source));
    } else {
      ownKeys(Object(source)).forEach(function (key) {
        Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key));
      });
    }
  }

  return target;
}

function getJoiValidationErrors(err) {
  var _err$details;

  if (err === null || err === void 0 ? void 0 : (_err$details = err.details) === null || _err$details === void 0 ? void 0 : _err$details.length) {
    const details = JSON.parse(JSON.stringify(err.details));
    const joiData = details.map(x => x.message.replace(new RegExp('"', "g"), ""));
    return joiData.join("; ");
  }

  return "";
}
function dateISOValidation({
  isRequired,
  defaultVal
} = {}) {
  if (isRequired === true) {
    if (defaultVal) {
      return Joi.string().isoDate().required().strict(false).default(defaultVal);
    }

    return Joi.string().isoDate().required().strict(false);
  }

  if (defaultVal) {
    return Joi.string().isoDate().strict(false).allow(null).default(defaultVal);
  }

  return Joi.string().isoDate().strict(false).empty("").allow(null).default(null);
}

class LoggingServiceBase {
  log(message, ...optionalParams) {
    console.log(message);
  }

}

const LoggingService = new LoggingServiceBase();

class UtilServiceBase {
  convertHexadecimalToNumber(hexString) {
    return parseInt(hexString, 16);
  }

  getRandomDecimal(min, max) {
    return Math.random() * (max - min) + min;
  }

  getRandomInt(min, max) {
    const minVal = Math.ceil(min);
    const maxVal = Math.floor(max);
    return Math.floor(Math.random() * (maxVal - minVal + 1)) + minVal;
  }

  getRandomString(count) {
    let text = "";
    const possible = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";

    for (let i = 0; i < count; i++) {
      const random = Math.floor(Math.random() * possible.length);
      text += possible[random];
    }

    return text;
  }

  getRandomAlphabet(count) {
    let txt = "";
    const alphabets = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

    for (let i = 0; i < count; i++) {
      const random = Math.floor(Math.random() * alphabets.length);
      txt += alphabets[random];
    }

    return txt;
  }

  camelCaseToSentenceCase(text) {
    const result = text.replace(/([A-Z])/g, " $1");
    const rst = result.charAt(0).toUpperCase() + result.slice(1);
    return rst.trim();
  }

  toTitleCase(text) {
    return text.replace(/\w\S*/g, txt => {
      return txt.charAt(0).toUpperCase() + txt.substr(1).toLowerCase();
    });
  }

  isValidPhoneNumber(_str) {
    if (isNaN(Number(_str))) {
      return false;
    }

    const isNum = /^\d+$/.test(_str);
    return isNum;
  }

  convertObjectPlainObject(objData) {
    const objDataPlain = JSON.parse(JSON.stringify(objData));
    return objDataPlain;
  }

  removeDuplicateString(items) {
    if (!Array.isArray(items)) {
      return [];
    }

    const unique = {};
    items.forEach(i => {
      if (!unique[i]) {
        unique[i] = true;
      }
    });
    return Object.keys(unique);
  }

  objectHasAnyProperty(obj) {
    if (obj && typeof obj === "object") {
      return Object.keys(obj).length > 0;
    }

    return false;
  }

  waitUntilMilliseconds(ms) {
    return new Promise((resolve, reject) => {
      setTimeout(() => {
        resolve();
      }, ms);
    });
  }

  waitUntilSeconds(seconds) {
    return new Promise(resolve => {
      setTimeout(() => {
        resolve();
      }, Math.round(seconds * 1000));
    });
  }

  waitUntilMunites(munites) {
    return new Promise(resolve => {
      setTimeout(() => {
        resolve();
      }, Math.round(munites * 60 * 1000));
    });
  }

  generateDynamoTableKey(date) {
    const _now = date ? new Date(date) : new Date();

    const key = [`${_now.getFullYear()}`, `${_now.getMonth() + 1}`.padStart(2, "0"), `${_now.getDate()}`.padStart(2, "0"), "-", `${_now.getHours()}`.padStart(2, "0"), `${_now.getMinutes()}`.padStart(2, "0"), `${_now.getSeconds()}`.padStart(2, "0"), "-", uuid.v4()];
    return key.join("");
  }

}

const UtilService = new UtilServiceBase();

const conditionKeyMap = {
  $eq: "=",
  $notEq: "<>",
  $lt: "<",
  $lte: "<=",
  $gt: ">",
  $gte: ">=",
  $exists: "",
  $notExists: "",
  $in: "",
  $between: "",
  $contains: "",
  $notContains: "",
  $beginsWith: ""
};

function hasQueryConditionKey(key) {
  const queryCondKeys = Object.keys(conditionKeyMap);
  return queryCondKeys.includes(key);
}

const getRandom = () => [Math.round(Math.random() * 99999), Math.round(Math.random() * 88888), Math.round(Math.random() * 99)].join("");

class DynamoFilterQueryOperation {
  operation__filterFieldExist({
    fieldName
  }) {
    const attrKeyHash = `#attrKey1${getRandom()}`.toLowerCase();
    const result = {
      xExpressionAttributeNames: {
        [attrKeyHash]: fieldName
      },
      xFilterExpression: `attribute_exists (${attrKeyHash})`
    };
    return result;
  }

  operation__filterFieldNotExist({
    fieldName
  }) {
    const attrKeyHash = `#attrKey2${getRandom()}`.toLowerCase();
    const result = {
      xExpressionAttributeNames: {
        [attrKeyHash]: fieldName
      },
      xFilterExpression: `attribute_not_exists (${attrKeyHash})`
    };
    return result;
  }

  __helperFilterBasic({
    fieldName,
    val,
    conditionExpr
  }) {
    const keyAttr = `:attr${fieldName}${getRandom()}`.toLowerCase();
    const attrKeyHash = `#attrKey3${getRandom()}`.toLowerCase();
    const result = {
      xExpressionAttributeValues: {
        [keyAttr]: val
      },
      xExpressionAttributeNames: {
        [attrKeyHash]: fieldName
      },
      xFilterExpression: [attrKeyHash, conditionExpr, keyAttr].join(" ")
    };
    return result;
  }

  operation__filterIn({
    fieldName,
    attrValues
  }) {
    const expressAttrVal = {};
    const expressAttrName = {};
    const filterExpress = [];

    const _attrKeyHash = `#attrKey4${getRandom()}`.toLowerCase();

    expressAttrName[_attrKeyHash] = fieldName;
    attrValues.forEach(item => {
      const keyAttr = `:attr${fieldName}${getRandom()}`.toLowerCase();
      expressAttrVal[keyAttr] = item;
      filterExpress.push(`${_attrKeyHash} = ${keyAttr}`);
    });

    const _filterExpression = filterExpress.join(" OR ").trim();

    const _filterExpressionValue = filterExpress.length > 1 ? `(${_filterExpression})` : _filterExpression;

    const result = {
      xExpressionAttributeValues: _objectSpread2({}, expressAttrVal),
      xExpressionAttributeNames: _objectSpread2({}, expressAttrName),
      xFilterExpression: _filterExpressionValue
    };
    return result;
  }

  operation__filterContains({
    fieldName,
    term
  }) {
    const attrKeyHash = `#attrKey5${getRandom()}`.toLowerCase();
    const keyAttr = `:attr${fieldName}${getRandom()}`.toLowerCase();
    const result = {
      xExpressionAttributeValues: {
        [keyAttr]: term
      },
      xExpressionAttributeNames: {
        [attrKeyHash]: fieldName
      },
      xFilterExpression: `contains (${attrKeyHash}, ${keyAttr})`
    };
    return result;
  }

  operation__filterBetween({
    fieldName,
    from,
    to
  }) {
    const _attrKeyHash = `#attrKey6${getRandom()}`.toLowerCase();

    const _fromKey = `:fromKey${getRandom()}`.toLowerCase();

    const _toKey = `:toKey${getRandom()}`.toLowerCase();

    const result = {
      xExpressionAttributeValues: {
        [_fromKey]: from,
        [_toKey]: to
      },
      xExpressionAttributeNames: {
        [_attrKeyHash]: fieldName
      },
      xFilterExpression: [_attrKeyHash, "between", _fromKey, "and", _toKey].join(" ")
    };
    return result;
  }

  operation__filterBeginsWith({
    fieldName,
    term
  }) {
    const _attrKeyHash = `#attrKey7${getRandom()}`.toLowerCase();

    const keyAttr = `:attr${fieldName}${getRandom()}`.toLowerCase();
    const result = {
      xExpressionAttributeValues: {
        [keyAttr]: term
      },
      xExpressionAttributeNames: {
        [_attrKeyHash]: fieldName
      },
      xFilterExpression: `begins_with (${_attrKeyHash}, ${keyAttr})`
    };
    return result;
  }

  operation__translateAdvancedQueryOperation({
    fieldName,
    queryObject
  }) {
    const queryConditions = [];
    Object.keys(queryObject).forEach(condKey => {
      const conditionKey = condKey;
      const _conditionObjValue = queryObject[conditionKey];

      if (conditionKey === "$between") {
        if (Array.isArray(_conditionObjValue)) {
          const _queryConditions = this.operation__filterBetween({
            fieldName: fieldName,
            from: _conditionObjValue[0],
            to: _conditionObjValue[1]
          });

          queryConditions.push(_queryConditions);
        }
      } else if (conditionKey === "$beginsWith") {
        const _queryConditions = this.operation__filterBeginsWith({
          fieldName: fieldName,
          term: _conditionObjValue
        });

        queryConditions.push(_queryConditions);
      } else if (conditionKey === "$contains") {
        const _queryConditions = this.operation__filterContains({
          fieldName: fieldName,
          term: _conditionObjValue
        });

        queryConditions.push(_queryConditions);
      } else if (conditionKey === "$in") {
        if (Array.isArray(_conditionObjValue)) {
          const _queryConditions = this.operation__filterIn({
            fieldName: fieldName,
            attrValues: _conditionObjValue
          });

          queryConditions.push(_queryConditions);
        }
      } else if (conditionKey === "$notContains") {
        const _queryConditions = this.operation__filterContains({
          fieldName: fieldName,
          term: _conditionObjValue
        });

        _queryConditions.xFilterExpression = `NOT ${_queryConditions.xFilterExpression}`;
        queryConditions.push(_queryConditions);
      } else if (conditionKey === "$exists") {
        const _queryConditions = this.operation__filterFieldExist({
          fieldName: fieldName
        });

        queryConditions.push(_queryConditions);
      } else if (conditionKey === "$notExists") {
        const _queryConditions = this.operation__filterFieldNotExist({
          fieldName: fieldName
        });

        queryConditions.push(_queryConditions);
      } else {
        if (hasQueryConditionKey(conditionKey)) {
          const conditionExpr = conditionKeyMap[conditionKey];

          if (conditionExpr) {
            const _queryConditions = this.__helperFilterBasic({
              fieldName: fieldName,
              val: _conditionObjValue,
              conditionExpr: conditionExpr
            });

            queryConditions.push(_queryConditions);
          }
        }
      }
    });
    return queryConditions;
  }

  operation_translateBasicQueryOperation({
    fieldName,
    queryObject
  }) {
    const _queryConditions = this.__helperFilterBasic({
      fieldName: fieldName,
      val: queryObject,
      conditionExpr: "="
    });

    return _queryConditions;
  }

  __helperDynamoFilterOperation({
    queryDefs,
    projectionFields
  }) {
    let queryAndConditions = [];
    let queryOrConditions = [];
    let _projectionExpression = undefined;
    const andFilterExpressionArray = [];
    const orFilterExpressionArray = [];
    let _expressionAttributeValues = {};
    let _expressionAttributeNames = {};
    Object.keys(queryDefs).forEach(fieldName_or_$or => {
      if (fieldName_or_$or === "$or") {
        const $orKey = fieldName_or_$or;
        const orArray = queryDefs[$orKey];

        if (Array.isArray(orArray)) {
          orArray.forEach(orQuery => {
            Object.keys(orQuery).forEach(fieldName => {
              const orQueryObjectOrValue = orQuery[fieldName];

              if (typeof orQueryObjectOrValue === "object") {
                const _orQueryCond = this.operation__translateAdvancedQueryOperation({
                  fieldName,
                  queryObject: orQueryObjectOrValue
                });

                queryOrConditions = [...queryOrConditions, ..._orQueryCond];
              } else {
                const _orQueryConditions = this.operation_translateBasicQueryOperation({
                  fieldName,
                  queryObject: orQueryObjectOrValue
                });

                queryOrConditions = [...queryOrConditions, _orQueryConditions];
              }
            });
          });
        }
      } else {
        const fieldName2 = fieldName_or_$or;
        const queryObjectOrValue = queryDefs[fieldName2];

        if (typeof queryObjectOrValue === "object") {
          const _queryCond = this.operation__translateAdvancedQueryOperation({
            fieldName: fieldName2,
            queryObject: queryObjectOrValue
          });

          queryAndConditions = [...queryAndConditions, ..._queryCond];
        } else {
          const _queryConditions = this.operation_translateBasicQueryOperation({
            fieldName: fieldName2,
            queryObject: queryObjectOrValue
          });

          queryAndConditions = [...queryAndConditions, _queryConditions];
        }
      }
    });

    for (const item of queryAndConditions) {
      _expressionAttributeNames = _objectSpread2(_objectSpread2({}, _expressionAttributeNames), item.xExpressionAttributeNames);
      _expressionAttributeValues = _objectSpread2(_objectSpread2({}, _expressionAttributeValues), item.xExpressionAttributeValues);
      andFilterExpressionArray.push(item.xFilterExpression);
    }

    for (const item2 of queryOrConditions) {
      _expressionAttributeNames = _objectSpread2(_objectSpread2({}, _expressionAttributeNames), item2.xExpressionAttributeNames);
      _expressionAttributeValues = _objectSpread2(_objectSpread2({}, _expressionAttributeValues), item2.xExpressionAttributeValues);
      orFilterExpressionArray.push(item2.xFilterExpression);
    }

    let _andfilterExpression = null;
    let _orfilterExpression = null;

    if (andFilterExpressionArray === null || andFilterExpressionArray === void 0 ? void 0 : andFilterExpressionArray.length) {
      _andfilterExpression = andFilterExpressionArray.join(" AND ").trim();
    }

    if (orFilterExpressionArray === null || orFilterExpressionArray === void 0 ? void 0 : orFilterExpressionArray.length) {
      _orfilterExpression = orFilterExpressionArray.join(" OR ").trim();
    }

    let _filterExpression = "";

    if (_andfilterExpression && _orfilterExpression) {
      _filterExpression = `(${_andfilterExpression}) AND (${_orfilterExpression})`;
    } else if (_andfilterExpression) {
      _filterExpression = _andfilterExpression;
    } else if (_orfilterExpression) {
      _filterExpression = _orfilterExpression;
    }

    if ((projectionFields === null || projectionFields === void 0 ? void 0 : projectionFields.length) && Array.isArray(projectionFields)) {
      const _projection_expressionAttributeNames = {};
      projectionFields.forEach(field => {
        if (typeof field === "string") {
          const attrKeyHash = `#attrKey8${getRandom()}`.toLowerCase();
          _projection_expressionAttributeNames[attrKeyHash] = field;
        }
      });
      _projectionExpression = Object.keys(_projection_expressionAttributeNames).join(", ");
      _expressionAttributeNames = _objectSpread2(_objectSpread2({}, _projection_expressionAttributeNames), _expressionAttributeNames);
    }

    const _expressionAttributeValuesFinal = UtilService.objectHasAnyProperty(_expressionAttributeValues) ? _expressionAttributeValues : undefined;

    const _expressionAttributeNamesFinal = UtilService.objectHasAnyProperty(_expressionAttributeNames) ? _expressionAttributeNames : undefined;

    const queryExpressions = {
      expressionAttributeValues: _expressionAttributeValuesFinal,
      filterExpression: _filterExpression,
      projectionExpressionAttr: _projectionExpression,
      expressionAttributeNames: _expressionAttributeNamesFinal
    };
    LoggingService.log({
      queryExpressions
    });
    return queryExpressions;
  }

}

class DynamoQueryScanProcessor {
  async __helperDynamoQueryProcessor({
    evaluationLimit,
    params,
    pageSize,
    lastKeyHash,
    orderDesc,
    hashKeyAndSortKey,
    dynamoDbClient
  }) {
    return await this.__helperDynamoQueryScanProcessor({
      dynamoDbClient,
      operation: "query",
      evaluationLimit,
      params,
      pageSize,
      lastKeyHash,
      orderDesc,
      hashKeyAndSortKey
    });
  }

  __helperDynamoQueryScanProcessor({
    operation,
    evaluationLimit,
    params,
    pageSize,
    lastKeyHash,
    orderDesc,
    hashKeyAndSortKey,
    dynamoDbClient
  }) {
    const xDefaultEvaluationLimit = 10;
    const xMinEvaluationLimit = 5;
    const xMaxEvaluationLimit = 500;
    console.log({
      processorParamsInit: {
        operation,
        pageSize,
        orderDesc,
        lastKeyHash,
        evaluationLimit,
        hashKeyAndSortKey,
        params
      }
    });
    return new Promise((resolve, reject) => {
      let returnedItems = [];
      let _evaluationLimit = 0;

      if (pageSize) {
        _evaluationLimit = xDefaultEvaluationLimit;

        if (evaluationLimit) {
          _evaluationLimit = evaluationLimit;
        }

        if (_evaluationLimit < xMinEvaluationLimit) {
          _evaluationLimit = xMinEvaluationLimit;
        } else if (_evaluationLimit > xMaxEvaluationLimit) {
          _evaluationLimit = xDefaultEvaluationLimit;
        }

        if (pageSize > _evaluationLimit) {
          _evaluationLimit = pageSize + 1;
        }
      }

      const queryScanUntilDone = (err, data) => {
        if (err) {
          var _returnedItems;

          console.log(err, err.stack);

          if ((_returnedItems = returnedItems) === null || _returnedItems === void 0 ? void 0 : _returnedItems.length) {
            resolve({
              mainResult: returnedItems
            });
          } else {
            reject(err.stack);
          }
        } else {
          var _data$Items;

          if (data === null || data === void 0 ? void 0 : (_data$Items = data.Items) === null || _data$Items === void 0 ? void 0 : _data$Items.length) {
            returnedItems = [...returnedItems, ...data.Items];
          }

          if (returnedItems.length && (hashKeyAndSortKey === null || hashKeyAndSortKey === void 0 ? void 0 : hashKeyAndSortKey.length)) {
            const dataObj = returnedItems.slice(-1)[0];

            const customLastEvaluationKey = this.__createCustomLastEvaluationKey(dataObj, hashKeyAndSortKey);

            console.log({
              customLastEvaluationKey
            });
          }

          if (pageSize && returnedItems.length >= pageSize) {
            const scanResult = {
              mainResult: returnedItems
            };

            if (data.LastEvaluatedKey) {
              const lastKeyHash = this.__encodeLastKey(data.LastEvaluatedKey);

              scanResult.lastKeyHash = lastKeyHash;
            }

            resolve(scanResult);
          } else if (data.LastEvaluatedKey) {
            const _paramsDef = _objectSpread2({}, params);

            _paramsDef.ExclusiveStartKey = data.LastEvaluatedKey;

            if (_evaluationLimit) {
              _paramsDef.Limit = _evaluationLimit;
            }

            console.log({
              operation,
              dynamoProcessorParams: _paramsDef
            });
            dynamoDbClient()[operation](_paramsDef, queryScanUntilDone);
          } else {
            resolve({
              mainResult: returnedItems
            });
          }
        }
      };

      const _params = _objectSpread2({}, params);

      if (_evaluationLimit) {
        _params.Limit = _evaluationLimit;
      }

      if (lastKeyHash) {
        const _lastEvaluatedKey = this.__decodeLastKey(lastKeyHash);

        if (_lastEvaluatedKey) {
          _params.ExclusiveStartKey = _lastEvaluatedKey;
        }
      }

      if (orderDesc === true && operation === "query") {
        _params["ScanIndexForward"] = false;
      }

      console.log({
        operation,
        dynamoProcessorParams: _params
      });
      dynamoDbClient()[operation](_params, queryScanUntilDone);
    });
  }

  __encodeLastKey(lastEvaluatedKey) {
    return Buffer.from(JSON.stringify(lastEvaluatedKey)).toString("base64");
  }

  __createCustomLastEvaluationKey(dataObj, primaryFieldNames) {
    const obj = {};
    primaryFieldNames.forEach(key => {
      if (typeof dataObj[key] !== "undefined") {
        obj[key] = dataObj[key];
      }
    });
    return Object.keys(obj).length > 0 ? obj : null;
  }

  __decodeLastKey(lastKeyHash) {
    let _lastEvaluatedKey;

    try {
      const _lastKeyHashStr = Buffer.from(lastKeyHash, "base64").toString();

      _lastEvaluatedKey = JSON.parse(_lastKeyHashStr);
    } catch (error) {
      _lastEvaluatedKey = undefined;
    }

    return _lastEvaluatedKey;
  }

}

function applyMixins(derivedCtor, baseCtors) {
  baseCtors.forEach(baseCtor => {
    Object.getOwnPropertyNames(baseCtor.prototype).forEach(name => {
      Object.defineProperty(derivedCtor.prototype, name, Object.getOwnPropertyDescriptor(baseCtor.prototype, name));
    });
  });
}

class DynamoFilterQueryMixin {
  allHelpValidateRequiredNumber(keyValueValidates) {
    const errors = [];
    Object.entries(keyValueValidates).forEach(([key, value]) => {
      if (!(!isNaN(Number(value)) && typeof value === "number")) {
        errors.push(`${key} is required`);
      }
    });

    if (errors.length) {
      throw new GenericDataError(`${errors.join("; ")}.`);
    }
  }

  allHelpCreateFriendlyError(message, statusCode) {
    return new GenericFriendlyError(message);
  }

  allHelpValidateRequiredString(keyValueValidates) {
    const errors = [];
    Object.entries(keyValueValidates).forEach(([key, value]) => {
      if (!(value && typeof value === "string")) {
        errors.push(`${key} is required`);
      }
    });

    if (errors.length) {
      throw new GenericDataError(`${errors.join("; ")}.`);
    }
  }

}
applyMixins(DynamoFilterQueryMixin, [DynamoFilterQueryOperation, DynamoQueryScanProcessor]);

const coreSchemaDefinition = {
  id: Joi.string().required().min(5).max(100),
  featureIdentity: Joi.string().required().min(3).max(50),
  lastModifierUserId: Joi.string().allow(null).empty("").default(null),
  lastModifiedDate: dateISOValidation(),
  creatorUserId: Joi.string().allow(null).empty("").default(null),
  createdAtDate: dateISOValidation({
    isRequired: true
  }),
  deleterUserId: Joi.string().allow(null).empty("").default(null),
  isDeleted: Joi.boolean().default(false)
};

class DynamoManageTable {
  constructor({
    dynamoDb,
    secondaryIndexOptions,
    tableFullName,
    partitionKeyFieldName,
    sortKeyFieldName
  }) {
    Object.defineProperty(this, "partitionKeyFieldName", {
      enumerable: true,
      configurable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "sortKeyFieldName", {
      enumerable: true,
      configurable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "dynamoDb", {
      enumerable: true,
      configurable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "tableFullName", {
      enumerable: true,
      configurable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "secondaryIndexOptions", {
      enumerable: true,
      configurable: true,
      writable: true,
      value: void 0
    });
    this.dynamoDb = dynamoDb;
    this.tableFullName = tableFullName;
    this.partitionKeyFieldName = partitionKeyFieldName;
    this.sortKeyFieldName = sortKeyFieldName;
    this.secondaryIndexOptions = secondaryIndexOptions;
  }

  _getLocalVariables() {
    return {
      partitionKeyFieldName: this.partitionKeyFieldName,
      sortKeyFieldName: this.sortKeyFieldName,
      tableFullName: this.tableFullName,
      secondaryIndexOptions: this.secondaryIndexOptions
    };
  }

  _dynamoDb() {
    return this.dynamoDb();
  }

  async allGetListOfTablesNamesOnlineBase() {
    const params = {
      Limit: 99
    };
    const listOfTables = await this._dynamoDb().listTables(params).promise();
    return listOfTables === null || listOfTables === void 0 ? void 0 : listOfTables.TableNames;
  }

  async allTableSettingUpdateTTLBase({
    attrName,
    isEnabled
  }) {
    const {
      tableFullName
    } = this._getLocalVariables();

    const params = {
      TableName: tableFullName,
      TimeToLiveSpecification: {
        AttributeName: attrName,
        Enabled: isEnabled
      }
    };
    const result = await this._dynamoDb().updateTimeToLive(params).promise();

    if (result === null || result === void 0 ? void 0 : result.TimeToLiveSpecification) {
      return result.TimeToLiveSpecification;
    }

    return null;
  }

  async allGetTableInfoBase() {
    try {
      var _result$Table;

      const {
        tableFullName
      } = this._getLocalVariables();

      const params = {
        TableName: tableFullName
      };
      const result = await this._dynamoDb().describeTable(params).promise();

      if ((result === null || result === void 0 ? void 0 : (_result$Table = result.Table) === null || _result$Table === void 0 ? void 0 : _result$Table.TableName) === tableFullName) {
        return result.Table;
      }

      return null;
    } catch (error) {
      LoggingService.log({
        "@allGetTableInfoBase": "",
        error: error === null || error === void 0 ? void 0 : error.message
      });
      return null;
    }
  }

  async allCheckTableExistsBase() {
    const result = await this.allGetTableInfoBase();

    if (!result) {
      return false;
    }

    if (result === null || result === void 0 ? void 0 : result.GlobalSecondaryIndexes) ;

    return true;
  }

  async _allUpdateGlobalSecondaryIndexBase({
    secondaryIndexOptions,
    existingTableInfo
  }) {
    try {
      var _existingTableInfo$Gl;

      const existingIndexNames = [];
      const staledIndexNames = [];
      const allIndexNames = [];
      const newSecondaryIndexOptions = [];
      const updateResults = [];

      if (existingTableInfo === null || existingTableInfo === void 0 ? void 0 : (_existingTableInfo$Gl = existingTableInfo.GlobalSecondaryIndexes) === null || _existingTableInfo$Gl === void 0 ? void 0 : _existingTableInfo$Gl.length) {
        existingTableInfo === null || existingTableInfo === void 0 ? void 0 : existingTableInfo.GlobalSecondaryIndexes.forEach(indexInfo => {
          if (indexInfo.IndexName) {
            existingIndexNames.push(indexInfo.IndexName);
          }
        });
      }

      secondaryIndexOptions === null || secondaryIndexOptions === void 0 ? void 0 : secondaryIndexOptions.forEach(newIndexInfo => {
        allIndexNames.push(newIndexInfo.indexName);
        const indexExists = existingIndexNames.includes(newIndexInfo.indexName);

        if (!indexExists) {
          newSecondaryIndexOptions.push(newIndexInfo);
        }
      });
      existingIndexNames.forEach(indexName => {
        const existsInList = allIndexNames.includes(indexName);

        if (!existsInList) {
          staledIndexNames.push(indexName);
        }
      });

      if (!(newSecondaryIndexOptions.length || staledIndexNames.length)) {
        return null;
      }

      let canUpdate = false;

      const {
        tableFullName
      } = this._getLocalVariables();

      LoggingService.log({
        secondaryIndexOptions: secondaryIndexOptions.length,
        newSecondaryIndexOptions: newSecondaryIndexOptions.length,
        staledIndexNames: staledIndexNames.length,
        tableName: tableFullName
      });

      if (newSecondaryIndexOptions.length) {
        canUpdate = true;
        let indexCount = 0;

        for (const indexOption of newSecondaryIndexOptions) {
          const params = {
            TableName: tableFullName,
            GlobalSecondaryIndexUpdates: []
          };
          indexCount++;

          const creationParams = this._getGlobalSecondaryIndexCreationParams({
            secondaryIndexOptions: [indexOption]
          });

          const indexName = creationParams.xGlobalSecondaryIndex[0].IndexName;
          params.AttributeDefinitions = [...creationParams.xAttributeDefinitions];
          creationParams.xGlobalSecondaryIndex.forEach(gsi => {
            var _params$GlobalSeconda;

            (_params$GlobalSeconda = params.GlobalSecondaryIndexUpdates) === null || _params$GlobalSeconda === void 0 ? void 0 : _params$GlobalSeconda.push({
              Create: _objectSpread2({}, gsi)
            });
          });
          const result = await this._dynamoDb().updateTable(params).promise();

          if (result === null || result === void 0 ? void 0 : result.TableDescription) {
            updateResults.push(result === null || result === void 0 ? void 0 : result.TableDescription);
          }

          LoggingService.log([`Creating Index: '${indexName}'`, `on table '${tableFullName}' started:`, new Date().toTimeString()].join(" "));

          if (indexCount !== newSecondaryIndexOptions.length) {
            const label = `Created Index '${indexName}'`;
            console.time(label);
            await UtilService.waitUntilMunites(5);
            console.timeEnd(label);
          }
        }
      }

      if (staledIndexNames.length) {
        if (canUpdate) {
          await UtilService.waitUntilMunites(4);
        }

        canUpdate = true;
        let indexCount = 0;

        for (const indexName of staledIndexNames) {
          var _params$GlobalSeconda2;

          const params = {
            TableName: tableFullName,
            GlobalSecondaryIndexUpdates: []
          };
          indexCount++;
          (_params$GlobalSeconda2 = params.GlobalSecondaryIndexUpdates) === null || _params$GlobalSeconda2 === void 0 ? void 0 : _params$GlobalSeconda2.push({
            Delete: {
              IndexName: indexName
            }
          });
          const result = await this._dynamoDb().updateTable(params).promise();

          if (result === null || result === void 0 ? void 0 : result.TableDescription) {
            updateResults.push(result === null || result === void 0 ? void 0 : result.TableDescription);
          }

          LoggingService.log([`Deleting Index: '${indexName}'`, `on table '${tableFullName}' started:`, new Date().toTimeString()].join(" "));

          if (indexCount !== staledIndexNames.length) {
            const label = `Deleted Index '${indexName}'`;
            console.time(label);
            await UtilService.waitUntilMunites(1);
            console.timeEnd(label);
          }
        }
      }

      if (!canUpdate) {
        return null;
      }

      if (updateResults.length) {
        console.log({
          "@allCreateGlobalSecondaryIndexBase": "",
          updateResults
        });
        return updateResults;
      }

      return null;
    } catch (error) {
      LoggingService.log({
        "@allCreateGlobalSecondaryIndexBase": "",
        error: error === null || error === void 0 ? void 0 : error.message
      });
      return null;
    }
  }

  async allCreateTableIfNotExistsBase() {
    const {
      secondaryIndexOptions
    } = this._getLocalVariables();

    const existingTableInfo = await this.allGetTableInfoBase();

    if (existingTableInfo) {
      var _existingTableInfo$Gl2;

      if (secondaryIndexOptions === null || secondaryIndexOptions === void 0 ? void 0 : secondaryIndexOptions.length) {
        await this._allUpdateGlobalSecondaryIndexBase({
          secondaryIndexOptions,
          existingTableInfo
        });
      } else if ((_existingTableInfo$Gl2 = existingTableInfo.GlobalSecondaryIndexes) === null || _existingTableInfo$Gl2 === void 0 ? void 0 : _existingTableInfo$Gl2.length) {
        await this._allUpdateGlobalSecondaryIndexBase({
          secondaryIndexOptions: [],
          existingTableInfo
        });
      }

      return null;
    }

    return await this.allCreateTableBase();
  }

  _getGlobalSecondaryIndexCreationParams({
    secondaryIndexOptions
  }) {
    const {
      tableFullName
    } = this._getLocalVariables();

    const params = {
      KeySchema: [],
      AttributeDefinitions: [],
      TableName: tableFullName,
      GlobalSecondaryIndexes: []
    };
    const attributeDefinitionsNameList = [];
    secondaryIndexOptions.forEach(sIndex => {
      var _projectionFields, _params$GlobalSeconda3;

      const {
        indexName,
        keyFieldName,
        sortFieldName,
        dataType,
        projectionFieldsInclude
      } = sIndex;
      const _keyFieldName = keyFieldName;
      const _sortFieldName = sortFieldName;

      if (!attributeDefinitionsNameList.includes(_keyFieldName)) {
        var _params$AttributeDefi;

        attributeDefinitionsNameList.push(_keyFieldName);
        params === null || params === void 0 ? void 0 : (_params$AttributeDefi = params.AttributeDefinitions) === null || _params$AttributeDefi === void 0 ? void 0 : _params$AttributeDefi.push({
          AttributeName: _keyFieldName,
          AttributeType: dataType
        });
      }

      if (!attributeDefinitionsNameList.includes(_sortFieldName)) {
        var _params$AttributeDefi2;

        attributeDefinitionsNameList.push(_sortFieldName);
        params === null || params === void 0 ? void 0 : (_params$AttributeDefi2 = params.AttributeDefinitions) === null || _params$AttributeDefi2 === void 0 ? void 0 : _params$AttributeDefi2.push({
          AttributeName: _sortFieldName,
          AttributeType: dataType
        });
      }

      let projectionFields = projectionFieldsInclude || [];
      let projectionType = "ALL";

      if ((_projectionFields = projectionFields) === null || _projectionFields === void 0 ? void 0 : _projectionFields.length) {
        projectionFields = projectionFields.filter(field => {
          return field !== _keyFieldName && field !== _sortFieldName;
        });

        if (projectionFields.length === 0) {
          projectionType = "KEYS_ONLY";
        } else {
          projectionType = "INCLUDE";
        }
      }

      (_params$GlobalSeconda3 = params.GlobalSecondaryIndexes) === null || _params$GlobalSeconda3 === void 0 ? void 0 : _params$GlobalSeconda3.push({
        IndexName: indexName,
        Projection: {
          ProjectionType: projectionType,
          NonKeyAttributes: projectionType === "INCLUDE" ? projectionFields : undefined
        },
        KeySchema: [{
          AttributeName: _keyFieldName,
          KeyType: "HASH"
        }, {
          AttributeName: _sortFieldName,
          KeyType: "RANGE"
        }]
      });
    });
    return {
      xAttributeDefinitions: params.AttributeDefinitions || [],
      xGlobalSecondaryIndex: params.GlobalSecondaryIndexes || []
    };
  }

  async allCreateTableBase() {
    const {
      partitionKeyFieldName,
      sortKeyFieldName,
      tableFullName,
      secondaryIndexOptions
    } = this._getLocalVariables();

    const params = {
      AttributeDefinitions: [{
        AttributeName: partitionKeyFieldName,
        AttributeType: "S"
      }, {
        AttributeName: sortKeyFieldName,
        AttributeType: "S"
      }],
      KeySchema: [{
        AttributeName: partitionKeyFieldName,
        KeyType: "HASH"
      }, {
        AttributeName: sortKeyFieldName,
        KeyType: "RANGE"
      }],
      BillingMode: "PAY_PER_REQUEST",
      TableName: tableFullName
    };
    params.Tags = [{
      Key: "tablePrefix",
      Value: tableFullName
    }, {
      Key: `DDBTableGroupKey-${tableFullName}`,
      Value: tableFullName
    }];

    if (secondaryIndexOptions === null || secondaryIndexOptions === void 0 ? void 0 : secondaryIndexOptions.length) {
      var _creationParams$xAttr;

      const creationParams = this._getGlobalSecondaryIndexCreationParams({
        secondaryIndexOptions
      });

      if ((_creationParams$xAttr = creationParams.xAttributeDefinitions) === null || _creationParams$xAttr === void 0 ? void 0 : _creationParams$xAttr.length) {
        const existAttrDefNames = params.AttributeDefinitions.map(def => def.AttributeName);
        creationParams.xAttributeDefinitions.forEach(def => {
          const alreadyDefined = existAttrDefNames.includes(def.AttributeName);

          if (!alreadyDefined) {
            params.AttributeDefinitions.push(def);
          }
        });
      }

      params.GlobalSecondaryIndexes = [...creationParams.xGlobalSecondaryIndex];
    }

    LoggingService.log({
      "@allCreateTableBase, table: ": tableFullName
    });
    const result = await this._dynamoDb().createTable(params).promise();

    if (result === null || result === void 0 ? void 0 : result.TableDescription) {
      LoggingService.log([`@allCreateTableBase,`, `Created table: '${result === null || result === void 0 ? void 0 : result.TableDescription.TableName}'`, new Date().toTimeString()].join(" "));
      return result.TableDescription;
    }

    return null;
  }

  async allDeleteGlobalSecondaryIndexBase(indexName) {
    const {
      tableFullName
    } = this._getLocalVariables();

    const params = {
      TableName: tableFullName,
      GlobalSecondaryIndexUpdates: [{
        Delete: {
          IndexName: indexName
        }
      }]
    };
    const result = await this._dynamoDb().updateTable(params).promise();

    if (result === null || result === void 0 ? void 0 : result.TableDescription) {
      return result.TableDescription;
    }

    return null;
  }

}

function createTenantSchema(schemaMapDef) {
  return Joi.object().keys(_objectSpread2(_objectSpread2({}, schemaMapDef), coreSchemaDefinition));
}

class DynamoDataOperation extends DynamoFilterQueryMixin {
  constructor({
    dynamoDb,
    schemaDef,
    dynamoDbClient,
    secondaryIndexOptions,
    featureIdentityValue,
    baseTableName,
    strictRequiredFields,
    dataKeyGenerator
  }) {
    super();
    Object.defineProperty(this, "here_partitionKeyFieldName", {
      enumerable: true,
      configurable: true,
      writable: true,
      value: "id"
    });
    Object.defineProperty(this, "here_sortKeyFieldName", {
      enumerable: true,
      configurable: true,
      writable: true,
      value: "featureIdentity"
    });
    Object.defineProperty(this, "here_dynamoDbClient", {
      enumerable: true,
      configurable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "here_dynamoDb", {
      enumerable: true,
      configurable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "here_dataKeyGenerator", {
      enumerable: true,
      configurable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "here_schema", {
      enumerable: true,
      configurable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "here_marshaller", {
      enumerable: true,
      configurable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "here_tableFullName", {
      enumerable: true,
      configurable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "here_strictRequiredFields", {
      enumerable: true,
      configurable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "here_featureIdentityValue", {
      enumerable: true,
      configurable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "here_secondaryIndexOptions", {
      enumerable: true,
      configurable: true,
      writable: true,
      value: void 0
    });
    Object.defineProperty(this, "here_tableManager", {
      enumerable: true,
      configurable: true,
      writable: true,
      value: void 0
    });
    this.here_dynamoDb = dynamoDb;
    this.here_dynamoDbClient = dynamoDbClient;
    this.here_dataKeyGenerator = dataKeyGenerator;
    this.here_schema = createTenantSchema(schemaDef);
    this.here_tableFullName = baseTableName;
    this.here_marshaller = new dynamodbAutoMarshaller.Marshaller({
      onEmpty: "omit"
    });
    this.here_featureIdentityValue = featureIdentityValue;
    this.here_secondaryIndexOptions = secondaryIndexOptions;
    this.here_strictRequiredFields = strictRequiredFields;
  }

  getTableManager() {
    if (!this.here_tableManager) {
      this.here_tableManager = new DynamoManageTable({
        dynamoDb: this.here_dynamoDb,
        secondaryIndexOptions: this.here_secondaryIndexOptions,
        tableFullName: this.here_tableFullName,
        partitionKeyFieldName: this.here_partitionKeyFieldName,
        sortKeyFieldName: this.here_sortKeyFieldName
      });
    }

    return this.here_tableManager;
  }

  _dynamoDbClient() {
    return this.here_dynamoDbClient();
  }

  _dynamoDb() {
    return this.here_dynamoDb();
  }

  generateDynamoTableKey() {
    return this.here_dataKeyGenerator();
  }

  _getLocalVariables() {
    return {
      partitionKeyFieldName: this.here_partitionKeyFieldName,
      sortKeyFieldName: this.here_sortKeyFieldName,
      featureIdentityValue: this.here_featureIdentityValue,
      tableFullName: this.here_tableFullName,
      secondaryIndexOptions: this.here_secondaryIndexOptions,
      strictRequiredFields: this.here_strictRequiredFields
    };
  }

  _getBaseObject({
    dataId
  }) {
    const {
      partitionKeyFieldName,
      sortKeyFieldName,
      featureIdentityValue
    } = this._getLocalVariables();

    const dataMust = {
      [partitionKeyFieldName]: dataId,
      [sortKeyFieldName]: featureIdentityValue
    };
    return dataMust;
  }

  _checkValidateMustBeAnObjectDataType(data) {
    if (!data || typeof data !== "object") {
      throw new GenericDataError(`Data MUST be valid object`);
    }
  }

  _checkValidateStrictRequiredFields(onDataObj) {
    this._checkValidateMustBeAnObjectDataType(onDataObj);

    const {
      strictRequiredFields
    } = this._getLocalVariables();

    if (strictRequiredFields === null || strictRequiredFields === void 0 ? void 0 : strictRequiredFields.length) {
      for (const field of strictRequiredFields) {
        if (onDataObj[field] === null || onDataObj[field] === undefined) {
          throw new GenericDataError(`Required field NOT defined`);
        }
      }
    }
  }

  async allCreateOneBase({
    data
  }) {
    this._checkValidateStrictRequiredFields(data);

    const {
      tableFullName,
      partitionKeyFieldName
    } = this._getLocalVariables();

    let dataId = data[partitionKeyFieldName];

    if (!dataId) {
      dataId = this.generateDynamoTableKey();
    }

    const baseData = {
      createdAtDate: new Date().toISOString()
    };

    const dataMust = this._getBaseObject({
      dataId
    });

    const fullData = _objectSpread2(_objectSpread2(_objectSpread2({}, data), baseData), dataMust);

    const {
      validatedData,
      marshalled
    } = await this._allHelpValidateMarshallAndGetValue(fullData);
    const params = {
      TableName: tableFullName,
      Item: marshalled
    };
    await this._dynamoDb().putItem(params).promise();
    const result = validatedData;
    return result;
  }

  withConditionPassed({
    item,
    withCondition
  }) {
    if (item && (withCondition === null || withCondition === void 0 ? void 0 : withCondition.length)) {
      const isPassed = withCondition.every(({
        field,
        equals
      }) => {
        return item[field] !== undefined && item[field] === equals;
      });
      return isPassed;
    }

    return true;
  }

  async allGetOneByIdBase({
    dataId,
    withCondition
  }) {
    const {
      partitionKeyFieldName,
      sortKeyFieldName,
      featureIdentityValue,
      tableFullName
    } = this._getLocalVariables();

    this.allHelpValidateRequiredString({
      QueryGetOnePartitionKey: dataId,
      QueryGetOneSortKey: featureIdentityValue
    });
    const params = {
      TableName: tableFullName,
      Key: {
        [partitionKeyFieldName]: dataId,
        [sortKeyFieldName]: featureIdentityValue
      }
    };
    const result = await this._dynamoDbClient().get(params).promise();
    const item = result.Item;

    if (!item) {
      return null;
    }

    const isPassed = this.withConditionPassed({
      withCondition,
      item
    });

    if (!isPassed) {
      return null;
    }

    return item;
  }

  async allUpdateOneDirectBase({
    data
  }) {
    this._checkValidateStrictRequiredFields(data);

    const {
      tableFullName,
      partitionKeyFieldName
    } = this._getLocalVariables();

    const dataId = data[partitionKeyFieldName];

    if (!dataId) {
      throw new GenericDataError("Update data requires sort key field value");
    }

    const dataMust = this._getBaseObject({
      dataId
    });

    const baseData = {
      lastModifiedDate: new Date().toISOString()
    };

    const fullData = _objectSpread2(_objectSpread2(_objectSpread2({}, data), baseData), dataMust);

    const {
      validatedData,
      marshalled
    } = await this._allHelpValidateMarshallAndGetValue(fullData);
    const params = {
      TableName: tableFullName,
      Item: marshalled
    };
    await this._dynamoDb().putItem(params).promise();
    const result = validatedData;
    return result;
  }

  async allUpdateOneByIdBase({
    dataId,
    data,
    withCondition
  }) {
    this._checkValidateStrictRequiredFields(data);

    const {
      tableFullName,
      partitionKeyFieldName
    } = this._getLocalVariables();

    this.allHelpValidateRequiredString({
      Update1DataId: dataId
    });
    const dataInDb = await this.allGetOneByIdBase({
      dataId
    });

    if (!(dataInDb && dataInDb[partitionKeyFieldName])) {
      throw new GenericDataError("Data does NOT exists");
    }

    const isPassed = this.withConditionPassed({
      withCondition,
      item: dataInDb
    });

    if (!isPassed) {
      throw new GenericDataError("Update condition failed");
    }

    const dataMust = this._getBaseObject({
      dataId: dataInDb[partitionKeyFieldName]
    });

    const fullData = _objectSpread2(_objectSpread2(_objectSpread2({}, dataInDb), data), dataMust);

    const {
      validatedData,
      marshalled
    } = await this._allHelpValidateMarshallAndGetValue(fullData);
    const params = {
      TableName: tableFullName,
      Item: marshalled
    };
    await this._dynamoDb().putItem(params).promise();
    const result = validatedData;
    return result;
  }

  async _allHelpValidateMarshallAndGetValue(data) {
    const {
      error,
      value
    } = this.here_schema.validate(data, {
      stripUnknown: true
    });

    if (error) {
      var _getJoiValidationErro;

      const msg = (_getJoiValidationErro = getJoiValidationErrors(error)) !== null && _getJoiValidationErro !== void 0 ? _getJoiValidationErro : "Validation error occured";
      throw new GenericDataError(msg);
    }

    const marshalledData = this.here_marshaller.marshallItem(value);
    return await Promise.resolve({
      validatedData: value,
      marshalled: marshalledData
    });
  }

  async allQueryGetManyByConditionBase(paramOptions) {
    var _result$mainResult;

    paramOptions.pagingParams = undefined;
    const result = await this.allQueryGetManyByConditionPaginateBase(paramOptions);

    if (result === null || result === void 0 ? void 0 : (_result$mainResult = result.mainResult) === null || _result$mainResult === void 0 ? void 0 : _result$mainResult.length) {
      return result.mainResult;
    }

    return [];
  }

  async allQueryGetManyByConditionPaginateBase(paramOptions) {
    var _paramOptions$partiti, _paramOptions$fields, _otherFilterExpressio, _paramOptions$pagingP;

    const {
      tableFullName,
      sortKeyFieldName,
      partitionKeyFieldName
    } = this._getLocalVariables();

    if (!(paramOptions === null || paramOptions === void 0 ? void 0 : (_paramOptions$partiti = paramOptions.partitionKeyQuery) === null || _paramOptions$partiti === void 0 ? void 0 : _paramOptions$partiti.equals) === undefined) {
      throw new GenericDataError("Invalid Hash key value");
    }

    if (!sortKeyFieldName) {
      throw new GenericDataError("Bad query sort configuration");
    }

    let sortKeyQuery = {};
    const sortKeyQueryData = paramOptions.sortKeyQuery;

    if (sortKeyQueryData) {
      if (sortKeyQueryData[sortKeyFieldName]) {
        sortKeyQuery = {
          [sortKeyFieldName]: sortKeyQueryData[sortKeyFieldName]
        };
      } else {
        throw new GenericDataError("Invalid Sort key value");
      }
    }

    const filterHashSortKey = this.__helperDynamoFilterOperation({
      queryDefs: _objectSpread2(_objectSpread2({}, sortKeyQuery), {
        [partitionKeyFieldName]: paramOptions.partitionKeyQuery.equals
      }),
      projectionFields: (_paramOptions$fields = paramOptions === null || paramOptions === void 0 ? void 0 : paramOptions.fields) !== null && _paramOptions$fields !== void 0 ? _paramOptions$fields : undefined
    });

    let otherFilterExpression = undefined;
    let otherExpressionAttributeValues = undefined;
    let otherExpressionAttributeNames = undefined;

    if (paramOptions === null || paramOptions === void 0 ? void 0 : paramOptions.query) {
      const filterOtherAttr = this.__helperDynamoFilterOperation({
        queryDefs: paramOptions.query,
        projectionFields: null
      });

      otherExpressionAttributeValues = filterOtherAttr.expressionAttributeValues;
      otherExpressionAttributeNames = filterOtherAttr.expressionAttributeNames;

      if ((filterOtherAttr === null || filterOtherAttr === void 0 ? void 0 : filterOtherAttr.filterExpression) && (filterOtherAttr === null || filterOtherAttr === void 0 ? void 0 : filterOtherAttr.filterExpression.length) > 1) {
        otherFilterExpression = filterOtherAttr.filterExpression;
      }
    }

    const params = {
      TableName: tableFullName,
      KeyConditionExpression: filterHashSortKey.filterExpression,
      ExpressionAttributeValues: _objectSpread2(_objectSpread2({}, otherExpressionAttributeValues), filterHashSortKey.expressionAttributeValues),
      FilterExpression: (_otherFilterExpressio = otherFilterExpression) !== null && _otherFilterExpressio !== void 0 ? _otherFilterExpressio : undefined,
      ExpressionAttributeNames: _objectSpread2(_objectSpread2({}, otherExpressionAttributeNames), filterHashSortKey.expressionAttributeNames)
    };

    if (filterHashSortKey === null || filterHashSortKey === void 0 ? void 0 : filterHashSortKey.projectionExpressionAttr) {
      params.ProjectionExpression = filterHashSortKey.projectionExpressionAttr;
    }

    if ((paramOptions === null || paramOptions === void 0 ? void 0 : (_paramOptions$pagingP = paramOptions.pagingParams) === null || _paramOptions$pagingP === void 0 ? void 0 : _paramOptions$pagingP.orderDesc) === true) {
      params.ScanIndexForward = false;
    }

    const hashKeyAndSortKey = [partitionKeyFieldName, sortKeyFieldName];

    const paginationObjects = _objectSpread2({}, paramOptions.pagingParams);

    const result = await this.__helperDynamoQueryProcessor(_objectSpread2({
      dynamoDbClient: () => this._dynamoDbClient(),
      params,
      hashKeyAndSortKey
    }, paginationObjects));
    return result;
  }

  async allBatchGetManyByIdsBase({
    dataIds,
    fields,
    withCondition
  }) {
    dataIds.forEach(dataId => {
      this.allHelpValidateRequiredString({
        BatchGetDataId: dataId
      });
    });

    const getRandom = () => ["rand", Math.round(Math.random() * 99999), Math.round(Math.random() * 88888), Math.round(Math.random() * 99)].join("");

    return new Promise((resolve, reject) => {
      const {
        tableFullName,
        partitionKeyFieldName,
        sortKeyFieldName,
        featureIdentityValue
      } = this._getLocalVariables();

      const getArray = dataIds.map(dataId => {
        const params01 = {
          [partitionKeyFieldName]: {
            S: dataId
          },
          [sortKeyFieldName]: {
            S: featureIdentityValue
          }
        };
        return params01;
      });
      let projectionExpression = undefined;
      let expressionAttributeNames = undefined;

      if (fields === null || fields === void 0 ? void 0 : fields.length) {
        var _Object$keys;

        const _fields = [...fields];

        if ((fields === null || fields === void 0 ? void 0 : fields.length) && (withCondition === null || withCondition === void 0 ? void 0 : withCondition.length)) {
          withCondition.forEach(condition => {
            if (!fields.includes(condition.field)) {
              _fields.push(condition.field);
            }
          });
        }

        expressionAttributeNames = {};

        _fields.forEach(fieldName => {
          if (typeof fieldName === "string") {
            if (expressionAttributeNames) {
              const attrKeyHash = `#attrKey${getRandom()}k`.toLowerCase();
              expressionAttributeNames[attrKeyHash] = fieldName;
            }
          }
        });

        if ((_Object$keys = Object.keys(expressionAttributeNames)) === null || _Object$keys === void 0 ? void 0 : _Object$keys.length) {
          projectionExpression = Object.keys(expressionAttributeNames).join(",");
        } else {
          projectionExpression = undefined;
          expressionAttributeNames = undefined;
        }
      }

      const params = {
        RequestItems: {
          [tableFullName]: {
            Keys: [...getArray],
            ConsistentRead: true,
            ProjectionExpression: projectionExpression,
            ExpressionAttributeNames: expressionAttributeNames
          }
        }
      };
      let returnedItems = [];

      const resolveItemResults = resultItems => {
        if ((resultItems === null || resultItems === void 0 ? void 0 : resultItems.length) && (withCondition === null || withCondition === void 0 ? void 0 : withCondition.length)) {
          return resultItems.filter(item => {
            return withCondition.every(condition => {
              return item[condition.field] === condition.equals;
            });
          });
        }

        return resultItems;
      };

      const batchGetUntilDone = (err, data) => {
        if (err) {
          var _returnedItems;

          if ((_returnedItems = returnedItems) === null || _returnedItems === void 0 ? void 0 : _returnedItems.length) {
            resolve(resolveItemResults(returnedItems));
          } else {
            reject(err.stack);
          }
        } else {
          if (data === null || data === void 0 ? void 0 : data.Responses) {
            const itemListRaw = data.Responses[tableFullName];
            const itemList = itemListRaw.map(item => {
              return this.here_marshaller.unmarshallItem(item);
            });
            returnedItems = [...returnedItems, ...itemList];
          }

          if (data === null || data === void 0 ? void 0 : data.UnprocessedKeys) {
            const _params = {
              RequestItems: data.UnprocessedKeys
            };
            console.log({
              dynamoBatchGetParams: _params
            });

            this._dynamoDb().batchGetItem(_params, batchGetUntilDone);
          } else {
            resolve(resolveItemResults(returnedItems));
          }
        }
      };

      this._dynamoDb().batchGetItem(params, batchGetUntilDone);
    });
  }

  async allQuerySecondaryIndexBase(paramOption) {
    paramOption.pagingParams = undefined;
    const result = await this.allQuerySecondaryIndexPaginateBase(paramOption);

    if (result === null || result === void 0 ? void 0 : result.mainResult) {
      return result.mainResult;
    }

    return [];
  }

  async allQuerySecondaryIndexPaginateBase(paramOption) {
    var _otherFilterExpressio2;

    const {
      tableFullName,
      secondaryIndexOptions
    } = this._getLocalVariables();

    if (!(secondaryIndexOptions === null || secondaryIndexOptions === void 0 ? void 0 : secondaryIndexOptions.length)) {
      throw new GenericDataError("Invalid secondary index definitions");
    }

    const {
      indexName,
      partitionKeyQuery,
      sortKeyQuery,
      fields,
      pagingParams,
      query
    } = paramOption;
    const secondaryIndex = secondaryIndexOptions.find(item => {
      return item.indexName === indexName;
    });

    if (!secondaryIndex) {
      throw new GenericDataError("Secondary index not named/defined");
    }

    const partitionKeyFieldName = secondaryIndex.keyFieldName;
    const sortKeyFieldName = secondaryIndex.sortFieldName;
    const partitionSortKeyQuery = sortKeyQuery ? _objectSpread2(_objectSpread2({}, {
      [sortKeyFieldName]: sortKeyQuery
    }), {
      [partitionKeyFieldName]: partitionKeyQuery.equals
    }) : {
      [partitionKeyFieldName]: partitionKeyQuery.equals
    };

    const {
      expressionAttributeValues,
      filterExpression,
      projectionExpressionAttr,
      expressionAttributeNames
    } = this.__helperDynamoFilterOperation({
      queryDefs: partitionSortKeyQuery,
      projectionFields: fields !== null && fields !== void 0 ? fields : undefined
    });

    let otherFilterExpression = undefined;
    let otherExpressionAttributeValues = undefined;
    let otherExpressionAttributeNames = undefined;

    if (query) {
      var _otherAttr$filterExpr;

      const otherAttr = this.__helperDynamoFilterOperation({
        queryDefs: query,
        projectionFields: null
      });

      otherExpressionAttributeValues = otherAttr.expressionAttributeValues;
      otherExpressionAttributeNames = otherAttr.expressionAttributeNames;

      if ((otherAttr === null || otherAttr === void 0 ? void 0 : (_otherAttr$filterExpr = otherAttr.filterExpression) === null || _otherAttr$filterExpr === void 0 ? void 0 : _otherAttr$filterExpr.length) && (otherAttr === null || otherAttr === void 0 ? void 0 : otherAttr.filterExpression.length) > 1) {
        otherFilterExpression = otherAttr.filterExpression;
      }
    }

    const params = {
      TableName: tableFullName,
      IndexName: indexName,
      KeyConditionExpression: filterExpression,
      ExpressionAttributeValues: _objectSpread2(_objectSpread2({}, otherExpressionAttributeValues), expressionAttributeValues),
      FilterExpression: (_otherFilterExpressio2 = otherFilterExpression) !== null && _otherFilterExpressio2 !== void 0 ? _otherFilterExpressio2 : undefined,
      ExpressionAttributeNames: _objectSpread2(_objectSpread2({}, otherExpressionAttributeNames), expressionAttributeNames)
    };
    const orderDesc = (pagingParams === null || pagingParams === void 0 ? void 0 : pagingParams.orderDesc) === true;

    if (orderDesc) {
      params.ScanIndexForward = false;
    }

    if (projectionExpressionAttr) {
      params.ProjectionExpression = projectionExpressionAttr;
    }

    const hashKeyAndSortKey = [partitionKeyFieldName, sortKeyFieldName];
    const result = await this.__helperDynamoQueryProcessor(_objectSpread2({
      dynamoDbClient: () => this._dynamoDbClient(),
      params,
      orderDesc,
      hashKeyAndSortKey
    }, pagingParams));
    return result;
  }

  async allDeleteByIdBase({
    dataId,
    withCondition
  }) {
    this.allHelpValidateRequiredString({
      Del1SortKey: dataId
    });

    const {
      tableFullName,
      partitionKeyFieldName,
      sortKeyFieldName,
      featureIdentityValue
    } = this._getLocalVariables();

    const dataExist = await this.allGetOneByIdBase({
      dataId,
      withCondition
    });

    if (!(dataExist && dataExist[partitionKeyFieldName])) {
      throw new GenericDataError("Record does NOT exists");
    }

    const params = {
      TableName: tableFullName,
      Key: {
        [partitionKeyFieldName]: dataId,
        [sortKeyFieldName]: featureIdentityValue
      }
    };
    await this._dynamoDbClient().delete(params).promise();
    return dataExist;
  }

  async allDeleteManyDangerouselyByIds({
    dataIds
  }) {
    dataIds.forEach(sortKeyValue => {
      this.allHelpValidateRequiredString({
        DelSortKey: sortKeyValue
      });
    });

    const {
      tableFullName,
      partitionKeyFieldName,
      sortKeyFieldName,
      featureIdentityValue
    } = this._getLocalVariables();

    const delArray = dataIds.map(dataId => {
      const params01 = {
        DeleteRequest: {
          Key: {
            [partitionKeyFieldName]: {
              S: dataId
            },
            [sortKeyFieldName]: {
              S: featureIdentityValue
            }
          }
        }
      };
      return params01;
    });
    const params = {
      RequestItems: {
        [tableFullName]: delArray
      }
    };
    await this._dynamoDb().batchWriteItem(params).promise();
    return true;
  }

}

const DynamoDataOperation$1 = DynamoDataOperation;

exports.DynamoDataOperation = DynamoDataOperation$1;
exports.GenericDataError = GenericDataError;
exports.default = DynamoDataOperation;
//# sourceMappingURL=index.js.map
